#include <stdio.h>
#include <stdlib.h>

#define VPCALC_SECTION	"GFX"
#define VPCALC_SUBSECT	"SDL_8X"

#include "../config.h"
#include "../include/types.h"
#include "../include/debug.h"

#include "../include/mem.h"

#include "../include/graphics.h"
#include "sdl.h"

#define WIDTH	128
#define HEIGHT	64
#define COLOR_LEVELS	9

typedef struct _graphics_o {
	int active;
	uint contrast;
	uint base_addr;
	int *refresh;

	uint depth, scale;
	uint32 fg_color, bg_color;

	uint8 memory[WIDTH*HEIGHT];

	mem_o mem;
	sdl_o sdl;
} _graphics_o;

graphics_o graphics_new(mem_o mem, uint depth, uint scale)
{
	_graphics_o *new;

	new = (_graphics_o *)malloc(sizeof(_graphics_o));
	if (new == NULL) {
		_err_report("Unable to allocate memory for graphics unit.");
		return NULL;
	}

	new->sdl = sdl_init(WIDTH*scale, HEIGHT*scale, depth);
	if (new->sdl == NULL) {
		free(new);
		return NULL;
	}

	new->mem = mem;

	new->active = 0;
	new->contrast = 0x0000;
	new->base_addr = 0xc000;
	new->refresh = NULL;

	new->depth = depth;
	new->scale = scale;
	new->fg_color = sdl_color(new->sdl, 0xff, 0xff, 0xff);
	new->bg_color = sdl_color(new->sdl, 0x00, 0x00, 0x00);

	return (graphics_o)new;
}

void graphics_del(graphics_o graphics)
{
	_graphics_o *_graphics = (_graphics_o *)graphics;

	sdl_close(_graphics->sdl);
	free(_graphics);
}

INTERNAL void _putpixel(sdl_o sdl, int x, int y, int scale, uint32 color)
{
	int i, j;

	x *= scale;
	y *= scale;

	if (scale <= 1) {
		sdl_putpixel(sdl, x, y, color);

		return;
	}

	if (scale == 2) {
		sdl_putpixel(sdl, x,   y,   color);
		sdl_putpixel(sdl, x+1, y,   color);

		sdl_putpixel(sdl, x,   y+1, color);
		sdl_putpixel(sdl, x+1, y+1, color);

		return;
	}

	if (scale == 3) {
		sdl_putpixel(sdl, x,   y,   color);
		sdl_putpixel(sdl, x+1, y,   color);
		sdl_putpixel(sdl, x+2, y,   color);

		sdl_putpixel(sdl, x,   y+1, color);
		sdl_putpixel(sdl, x+1, y+1, color);
		sdl_putpixel(sdl, x+2, y+1, color);

		sdl_putpixel(sdl, x,   y+2, color);
		sdl_putpixel(sdl, x+1, y+2, color);
		sdl_putpixel(sdl, x+2, y+2, color);

		return;
	}

	if (scale == 4) {
		sdl_putpixel(sdl, x,   y,   color);
		sdl_putpixel(sdl, x+1, y,   color);
		sdl_putpixel(sdl, x+2, y,   color);
		sdl_putpixel(sdl, x+3, y,   color);

		sdl_putpixel(sdl, x,   y+1, color);
		sdl_putpixel(sdl, x+1, y+1, color);
		sdl_putpixel(sdl, x+2, y+1, color);
		sdl_putpixel(sdl, x+3, y+1, color);

		sdl_putpixel(sdl, x,   y+2, color);
		sdl_putpixel(sdl, x+1, y+2, color);
		sdl_putpixel(sdl, x+2, y+2, color);
		sdl_putpixel(sdl, x+3, y+2, color);

		sdl_putpixel(sdl, x,   y+3, color);
		sdl_putpixel(sdl, x+1, y+3, color);
		sdl_putpixel(sdl, x+2, y+3, color);
		sdl_putpixel(sdl, x+3, y+3, color);

		return;
	}

	for (i=0; i<scale; i++) {
		for (j=0; j<scale; j++) {
			sdl_putpixel(sdl, x+i, y+j, color);
		}
	}
}

void graphics_update(graphics_o graphics)
{
	_graphics_o *_graphics = (_graphics_o *)graphics;
	int i, p;
	uint len;
	uint16 addr;
	uint8 c, *color;

	color = _graphics->memory;

	len = (WIDTH*HEIGHT)/8;
	for (p=0; p<len; p++) {
		addr = p + _graphics->base_addr;

		c = mem_read_byte(_graphics->mem, addr);

		for (i=0; i<8; i++, c <<= 1) {
			if(c & 0x80) {
				(*color)++;
			}

			color++;
		}
	}
}

void graphics_draw(graphics_o graphics)
{
	_graphics_o *_graphics = (_graphics_o *)graphics;
	int i, j, k;
	int p = 0, len;
	uint32 clr;
	uint16 addr;
	uint8 c;

	if (!_graphics->active) {
		sdl_screenclr(_graphics->sdl, _graphics->bg_color);
		sdl_drawscreen(_graphics->sdl);
		return;
	}

	i = j = 0;
	len = (HEIGHT * WIDTH) / 8;
	for (p=0; p<len; p++) {
		addr = p + _graphics->base_addr;

		c = mem_read_byte(_graphics->mem, addr);

		for (k=0; k<8; k++, c <<= 1) {
			if (c & 0x80) {
				clr = _graphics->fg_color;
			} else {
				clr = _graphics->bg_color;
			}
			_putpixel(_graphics->sdl, i, j, _graphics->scale, clr);

			i++;
			if(i >= WIDTH) {
				i -= WIDTH;
				j++;
			}

			//if(j >= HEIGHT)
				//j -= HEIGHT;
		}
	}

	sdl_drawscreen(_graphics->sdl);
}

void graphics_palette(graphics_o graphics, int i, uint8 r, uint8 g, uint8 b)
{
	_graphics_o *_graphics = (_graphics_o *)graphics;

	if (i) {
		_graphics->fg_color = sdl_color(_graphics->sdl, r, g, b);

	} else {
		_graphics->bg_color = sdl_color(_graphics->sdl, r, g, b);
	}
}

void graphics_activate(graphics_o graphics, int value)
{
	_graphics_o *_graphics = (_graphics_o *)graphics;

	_graphics->active = value;
}

void graphics_set_contrast(graphics_o graphics, uint value)
{
	_graphics_o *_graphics = (_graphics_o *)graphics;

	_graphics->contrast = value;
}

void graphics_set_screen_addr(graphics_o graphics, uint scr_addr)
{
	_graphics_o *_graphics = (_graphics_o *)graphics;

	_report("screen_addr: $%04X", scr_addr);
	_graphics->base_addr = scr_addr;
}

void graphics_map_refresh(graphics_o graphics, int *refreshing)
{
	_graphics_o *_graphics = (_graphics_o *)graphics;

	_graphics->refresh = refreshing;
}

void graphics_write(graphics_o graphics, uint index, uint data)
{
	_graphics_o *_graphics = (_graphics_o *)graphics;

	index &= 0x03ff;

	mem_write(_graphics->mem, index + _graphics->base_addr, data);
}

uint graphics_read(graphics_o graphics, uint index)
{
	_graphics_o *_graphics = (_graphics_o *)graphics;

	index &= 0x03ff;

	return mem_read(_graphics->mem, index + _graphics->base_addr);
}
