#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/uio.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>

#define VPCALC_SECTION	"MEM"
#define VPCALC_SUBSECT	"86"

#include "../config.h"
#include "../include/types.h"
#include "../include/debug.h"

#include "../include/memory.h"

#define PAGE(p)		((p) << 14)
#define PAGE_SIZE	(1 << 14)
#define SCRATCH		(0x18)
#define PAGE_LAST	(0x19)

#define MEM_SIZE	(PAGE_SIZE * PAGE_LAST)
#define ADDR_MAX	(0x10000)

#define PROT_RAM	(PROT_READ | PROT_WRITE)
#define PROT_ROM	(PROT_READ)

#define FLAGS_STD	(MAP_FILE | MAP_SHARED)
#define FLAGS_FXD	(MAP_FILE | MAP_SHARED | MAP_FIXED)

typedef struct _memory_o {
	int mem_fd;
	uint8 *memory;

	uint8 *rd_mem, *wr_mem;

	uint pages[4];
	uint8* rblocks[4];
	uint8* wblocks[4];
} _memory_o;

static void _remap(uint8 *base, uint page, uint addr, int perms, int fd)
{
	void *ptr;

	page = PAGE(page);
	addr = PAGE(addr);

	ptr = mmap(base+page, PAGE_SIZE, perms, FLAGS_FXD, fd, addr);
	if (ptr == (void *)-1) {
		_err_report("Remap failure!");
		exit(1);
	}
}

memory_o memory_new(void)
{
	_memory_o *new;
	char zero_page[PAGE_SIZE];
	int i;

	new = (_memory_o *) malloc(sizeof(_memory_o));
	if (new == NULL) {
		_err_report("Out of memory!");
		return NULL;
	}

	memset(new, 0, sizeof(_memory_o));
	memset(zero_page, 0, PAGE_SIZE);

	new->mem_fd = open("blah.sta", O_RDWR | O_CREAT, 0644);
	if (new->mem_fd < 0) {
		_err_report("Unable to open mmap file");
		free(new);
		return NULL;
	}

	lseek(new->mem_fd, 0, SEEK_SET);
	for (i=0; i<PAGE_LAST; i++) {
		write(new->mem_fd, zero_page, PAGE_SIZE);
	}

	new->memory = mmap(NULL, MEM_SIZE, PROT_RAM, FLAGS_STD, new->mem_fd, 0);
	if (new->rd_mem == (void *)-1) {
		_err_report("read mmap failed");
		close(new->mem_fd);
		free(new);
		return NULL;
	}

	new->rd_mem = mmap(NULL, ADDR_MAX, PROT_ROM, FLAGS_STD, new->mem_fd, 0);
	if (new->rd_mem == (void *)-1) {
		_err_report("raw read mmap failed");
		munmap(new->memory, MEM_SIZE);
		close(new->mem_fd);
		free(new);
		return NULL;
	}

	new->wr_mem = mmap(NULL, ADDR_MAX, PROT_ROM, FLAGS_STD, new->mem_fd, 0);
	if (new->wr_mem == (void *)-1) {
		_err_report("write mmap failed");
		munmap(new->memory, MEM_SIZE);
		munmap(new->rd_mem, ADDR_MAX);
		close(new->mem_fd);
		free(new);
		return NULL;
	}

	new->pages[0] = 0x00;
	new->pages[1] = 0x01;
	new->pages[2] = 0x00;
	new->pages[3] = 0x10;

	new->rblocks[0] = new->memory + PAGE(new->pages[0]);
	new->rblocks[1] = new->memory + PAGE(new->pages[1] - 1);
	new->rblocks[2] = new->memory + PAGE(new->pages[2] - 2);
	new->rblocks[3] = new->memory + PAGE(new->pages[3] - 3);

	_remap(new->rd_mem, 1, new->pages[1], PROT_ROM, new->mem_fd);
	_remap(new->rd_mem, 2, new->pages[2], PROT_ROM, new->mem_fd);
	_remap(new->rd_mem, 3, new->pages[3], PROT_ROM, new->mem_fd);

	new->wblocks[0] = new->memory + PAGE(SCRATCH);
	new->wblocks[1] = new->memory + PAGE(SCRATCH - 1);
	new->wblocks[2] = new->memory + PAGE(SCRATCH - 2);
	new->wblocks[3] = new->memory + PAGE(new->pages[3] - 3);

	_remap(new->wr_mem, 0, SCRATCH, PROT_RAM, new->mem_fd);
	_remap(new->wr_mem, 1, SCRATCH, PROT_RAM, new->mem_fd);
	_remap(new->wr_mem, 2, SCRATCH, PROT_RAM, new->mem_fd);
	_remap(new->wr_mem, 3, new->pages[3], PROT_RAM, new->mem_fd);

	return (memory_o)new;
}

void memory_del(memory_o memory)
{
	_memory_o *_memory = (_memory_o *)memory;

	msync(_memory->memory, MEM_SIZE, MS_SYNC);

	munmap(_memory->memory, MEM_SIZE);
	munmap(_memory->rd_mem, ADDR_MAX);
	munmap(_memory->wr_mem, ADDR_MAX);

	close(_memory->mem_fd);

	free(_memory);
}

size_t memory_size(memory_o memory)
{
	return MEM_SIZE;
}

size_t memory_bus_size(memory_o memory)
{
	return 8;
}

size_t memory_addr_size(memory_o memory)
{
	return 16;
}

void memory_putimage(memory_o memory, void *data)
{
	_memory_o *_memory = (_memory_o *)memory;

	memcpy(_memory->memory, data, MEM_SIZE);
}

void memory_getimage(memory_o memory, void *data)
{
	_memory_o *_memory = (_memory_o *)memory;

	memcpy(data, _memory->memory, MEM_SIZE);
}

uint memory_fetch(memory_o memory, uint addr)
{
	_memory_o *_memory = (_memory_o *)memory;

	addr &= 0xFFFF;

	//return _memory->rblocks[addr >> 14][addr];
	return _memory->rd_mem[addr];
}

uint memory_read(memory_o memory, uint addr)
{
	_memory_o *_memory = (_memory_o *)memory;

	addr &= 0xFFFF;

	//return _memory->rblocks[addr >> 14][addr];
	return _memory->rd_mem[addr];
}

void memory_write(memory_o memory, uint addr, uint data)
{
	_memory_o *_memory = (_memory_o *)memory;

	addr &= 0xFFFF;

	//_memory->wblocks[addr >> 14][addr] = data;
	_memory->wr_mem[addr] = data;
}

uint8 memory_fetch_byte(memory_o memory, uint addr)
{
	_memory_o *_memory = (_memory_o *)memory;

	addr &= 0xFFFF;

	//return _memory->rblocks[addr >> 14][addr];
	return _memory->rd_mem[addr];
}

uint8 memory_read_byte(memory_o memory, uint addr)
{
	_memory_o *_memory = (_memory_o *)memory;

	addr &= 0xFFFF;

	//return _memory->rblocks[addr >> 14][addr];
	return _memory->rd_mem[addr];
}

void memory_write_byte(memory_o memory, uint addr, uint8 data)
{
	_memory_o *_memory = (_memory_o *)memory;

	addr &= 0xFFFF;

	//_memory->wblocks[addr >> 14][addr] = data;
	_memory->wr_mem[addr] = data;
}

void memory_remap(memory_o memory, uint page, uint addr, uint perm)
{
	_memory_o *_memory = (_memory_o *)memory;

	page &= 0x03;
	addr &= 0x1f;
	if (addr & 0x10)
		addr &= 0x17;

	_memory->pages[page] = addr;

	_memory->rblocks[page] = _memory->memory + PAGE(addr - page);
	_remap(_memory->rd_mem, page, addr, PROT_ROM, _memory->mem_fd);

	if (!(perm & PERM_WRITE))
		addr = SCRATCH;

	_memory->wblocks[page] = _memory->memory + PAGE(addr - page);
	_remap(_memory->wr_mem, page, addr, PROT_RAM, _memory->mem_fd);
}

int memory_raw(memory_o memory, void **rd_ptr, void **wr_ptr, void **ex_ptr)
{
	_memory_o *_memory = (_memory_o *)memory;

	*rd_ptr = _memory->rd_mem;
	*wr_ptr = _memory->wr_mem;
	*ex_ptr = _memory->rd_mem;

	return 1;
}
