#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>

#define VPCALC_SECTION	"LIB"
#define VPCALC_SUBSECT	"GFX"

#include "../config.h"
#include "../include/types.h"
#include "../include/debug.h"

#include "../include/mem.h"
#include "../include/gfx.h"

#include "../include/graphics.h"

typedef struct _gfx_o {
	void *library;

	graphics_o graphics;

	void (*del)(graphics_o graphics);

	void (*update)(graphics_o graphics);
	void (*draw)(graphics_o graphics);
	void (*palette)(graphics_o graphics, int i, uint8 r, uint8 g, uint8 b);

	void (*activate)(graphics_o graphics, int value);
	void (*set_contrast)(graphics_o graphics, uint value);
	void (*set_screen_addr)(graphics_o graphics, uint value);
	void (*map_refresh)(graphics_o graphics, int *refreshing);

	void (*write)(graphics_o graphics, uint index, uint data);
	uint (*read)(graphics_o graphics, uint index);
} _gfx_o;

gfx_o gfx_new(const char *libname, mem_o mem, uint depth, uint scale)
{
	_gfx_o *new;
	graphics_o (*graphics_new)(mem_o, uint depth, uint scale);
	const char *error;
  
	new = (_gfx_o *) malloc(sizeof(_gfx_o));
	if (new == NULL) {
		_err_report("Unable to allocate cpu structure.");

		return NULL;
	}

	new->library = dlopen(libname, RTLD_NOW);
	if (!new->library) {
		error = dlerror();
		_err_report(error);
		free(new);
		return NULL;
	}

	graphics_new = dlsym(new->library, "graphics_new");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->del = dlsym(new->library, "graphics_del");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->update = dlsym(new->library, "graphics_update");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->draw = dlsym(new->library, "graphics_draw");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->palette = dlsym(new->library, "graphics_palette");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->activate = dlsym(new->library, "graphics_activate");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->set_contrast = dlsym(new->library, "graphics_set_contrast");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->set_screen_addr = dlsym(new->library, "graphics_set_screen_addr");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->map_refresh = dlsym(new->library, "graphics_map_refresh");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->write = dlsym(new->library, "graphics_write");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->read = dlsym(new->library, "graphics_read");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->graphics = graphics_new(mem, depth, scale);
	if (new->graphics == NULL) {
		_err_report("failed to make new graphics unit.");

		dlclose(new->library);
		free(new);
		return NULL;
	}

	return (gfx_o)new;
}

void gfx_del(gfx_o gfx)
{
	_gfx_o *_gfx = (_gfx_o *)gfx;

	_gfx->del(_gfx->graphics);

	dlclose(_gfx->library);
	free(_gfx);
}

void gfx_update(gfx_o gfx)
{
	_gfx_o *_gfx = (_gfx_o *)gfx;

	_gfx->update(_gfx->graphics);
}

void gfx_draw(gfx_o gfx)
{
	_gfx_o *_gfx = (_gfx_o *)gfx;

	_gfx->draw(_gfx->graphics);
}

void gfx_palette(gfx_o gfx, int i, uint8 r, uint8 g, uint8 b)
{
	_gfx_o *_gfx = (_gfx_o *)gfx;

	_gfx->palette(_gfx->graphics, i, r, g, b);
}

void gfx_activate(gfx_o gfx, int value)
{
	_gfx_o *_gfx = (_gfx_o *)gfx;

	_gfx->activate(_gfx->graphics, value);
}

void gfx_set_contrast(gfx_o gfx, uint value)
{
	_gfx_o *_gfx = (_gfx_o *)gfx;

	_gfx->set_contrast(_gfx->graphics, value);
}

void gfx_set_screen_addr(gfx_o gfx, uint scr_addr)
{
	_gfx_o *_gfx = (_gfx_o *)gfx;

	_gfx->set_screen_addr(_gfx->graphics, scr_addr);
}

void gfx_map_refresh(gfx_o gfx, int *refreshing)
{
	_gfx_o *_gfx = (_gfx_o *)gfx;

	_gfx->map_refresh(_gfx->graphics, refreshing);
}

void gfx_write(gfx_o gfx, uint index, uint data)
{
	_gfx_o *_gfx = (_gfx_o *)gfx;

	_gfx->write(_gfx->graphics, index, data);
}

uint gfx_read(gfx_o gfx, uint index)
{
	_gfx_o *_gfx = (_gfx_o *)gfx;

	return _gfx->read(_gfx->graphics, index);
}
